from fastai.vision.data import ImageList
import pandas as pd
import json
from pathlib import Path
import random
import types
from textwrap import wrap
import sys
from typing import List, Callable, Tuple, Dict, cast
import warnings
import traceback
import xml.etree.ElementTree as ET

from traitlets.traitlets import parse_notifier_name
from .env import raise_fastai_import_error

try:
    import matplotlib.pyplot as plt
    from torch.utils.data import Dataset, DataLoader
    from fastai.text import Tokenizer, Vocab
    import fastai
    from fastai.vision import (
        get_transforms,
        open_image,
        image2np,
        Image,
        imagenet_stats,
        normalize,
        denormalize,
    )
    from fastai.vision.transform import crop
    from fastai.data_block import DataBunch
    from .pointcloud_data import get_device
    from .._utils.common import get_nbatches, get_top_padding
    import matplotlib.patheffects as PathEffects
    import torch

    HAS_FASTAI = True
except ImportError:
    import_exception = traceback.format_exc()
    HAS_FASTAI = False

StateType = Dict[str, torch.Tensor]
StepFunctionType = Callable[
    [torch.Tensor, StateType, int], Tuple[torch.Tensor, StateType]
]
StepFunctionTypeNoTimestep = Callable[
    [torch.Tensor, StateType], Tuple[torch.Tensor, StateType]
]


# Modified BeamSearch from AllenNLP library.
class BeamSearchAttention:
    """
    Implements the beam search algorithm for decoding the most likely sequences.
    [0]: https://arxiv.org/abs/1702.01806
    # Parameters
    end_index : `int`
        The index of the "stop" or "end" token in the target vocabulary.
    max_steps : `int`, optional (default = `50`)
        The maximum number of decoding steps to take, i.e. the maximum length
        of the predicted sequences.
    beam_size : `int`, optional (default = `10`)
        The width of the beam used.
    per_node_beam_size : `int`, optional (default = `beam_size`)
        The maximum number of candidates to consider per node, at each step in the search.
        If not given, this just defaults to `beam_size`. Setting this parameter
        to a number smaller than `beam_size` may give better results, as it can introduce
        more diversity into the search. See [Beam Search Strategies for Neural Machine Translation.
        Freitag and Al-Onaizan, 2017][0].
    """

    def __init__(
        self,
        end_index: int,
        max_steps: int = 50,
        beam_size: int = 10,
        per_node_beam_size: int = None,
    ) -> None:
        self._end_index = end_index
        self.max_steps = max_steps
        self.beam_size = beam_size
        self.per_node_beam_size = per_node_beam_size or beam_size

    @staticmethod
    def reconstruct_sequences(predictions, backpointers):
        # Reconstruct the sequences.
        # shape: [(batch_size, beam_size, 1)]
        reconstructed_predictions = [predictions[-1].unsqueeze(2)]

        # shape: (batch_size, beam_size)
        cur_backpointers = backpointers[-1]

        for timestep in range(len(predictions) - 2, 0, -1):
            # shape: (batch_size, beam_size, 1)
            cur_preds = (
                predictions[timestep].gather(1, cur_backpointers.long()).unsqueeze(2)
            )

            reconstructed_predictions.append(cur_preds)

            # shape: (batch_size, beam_size)
            cur_backpointers = backpointers[timestep - 1].gather(
                1, cur_backpointers.long()
            )

        # shape: (batch_size, beam_size, 1)
        final_preds = predictions[0].gather(1, cur_backpointers.long()).unsqueeze(2)

        reconstructed_predictions.append(final_preds)

        return reconstructed_predictions

    @torch.no_grad()
    def search(
        self,
        start_predictions: torch.Tensor,
        start_state: StateType,
        step: StepFunctionType,
    ) -> Tuple[torch.Tensor, torch.Tensor]:
        """
        Given a starting state and a step function, apply beam search to find the
        most likely target sequences.
        Notes
        -----
        If your step function returns `-inf` for some log probabilities
        (like if you're using a masked log-softmax) then some of the "best"
        sequences returned may also have `-inf` log probability. Specifically
        this happens when the beam size is smaller than the number of actions
        with finite log probability (non-zero probability) returned by the step function.
        Therefore if you're using a mask you may want to check the results from `search`
        and potentially discard sequences with non-finite log probability.
        # Parameters
        start_predictions : `torch.Tensor`
            A tensor containing the initial predictions with shape `(batch_size,)`.
            Usually the initial predictions are just the index of the "start" token
            in the target vocabulary.
        start_state : `StateType`
            The initial state passed to the `step` function. Each value of the state dict
            should be a tensor of shape `(batch_size, *)`, where `*` means any other
            number of dimensions.
        step : `StepFunctionType`
            A function that is responsible for computing the next most likely tokens,
            given the current state and the predictions from the last time step.
            The function should accept two arguments. The first being a tensor
            of shape `(group_size,)`, representing the index of the predicted
            tokens from the last time step, and the second being the current state.
            The `group_size` will be `batch_size * beam_size`, except in the initial
            step, for which it will just be `batch_size`.
            The function is expected to return a tuple, where the first element
            is a tensor of shape `(group_size, target_vocab_size)` containing
            the log probabilities of the tokens for the next step, and the second
            element is the updated state. The tensor in the state should have shape
            `(group_size, *)`, where `*` means any other number of dimensions.
        # Returns
        `Tuple[torch.Tensor, torch.Tensor]`
            Tuple of `(predictions, log_probabilities)`, where `predictions`
            has shape `(batch_size, beam_size, max_steps)` and `log_probabilities`
            has shape `(batch_size, beam_size)`.
        """

        # If the step function we're given does not take the time step argument, wrap it
        # in one that does.
        from inspect import signature

        step_signature = signature(step)
        if len(step_signature.parameters) < 3:
            old_step = cast(StepFunctionTypeNoTimestep, step)

            def new_step(
                last_predictions: torch.Tensor,
                state: Dict[str, torch.Tensor],
                time_step: int,
            ):
                return old_step(last_predictions, state)

            step = new_step

        batch_size = start_predictions.size()[0]

        # List of (batch_size, beam_size) tensors. One for each time step. Does not
        # include the start symbols, which are implicit.
        predictions: List[torch.Tensor] = []

        # List of (batch_size, beam_size) tensors. One for each time step. None for
        # the first.  Stores the index n for the parent prediction, i.e.
        # predictions[t-1][i][n], that it came from.
        backpointers: List[torch.Tensor] = []
        attention_maps = []

        # Calculate the first timestep. This is done outside the main loop
        # because we are going from a single decoder input (the output from the
        # encoder) to the top `beam_size` decoder outputs. On the other hand,
        # within the main loop we are going from the `beam_size` elements of the
        # beam to `beam_size`^2 candidates from which we will select the top
        # `beam_size` elements for the next iteration.
        # shape: (batch_size, num_classes)
        start_class_log_probabilities, state = step(start_predictions, start_state, 0)

        num_classes = start_class_log_probabilities.size()[1]

        # Make sure `per_node_beam_size` is not larger than `num_classes`.
        if self.per_node_beam_size > num_classes:
            raise Exception(
                f"Target vocab size ({num_classes:d}) too small "
                f"relative to per_node_beam_size ({self.per_node_beam_size:d}).\n"
                f"Please decrease beam_size or per_node_beam_size."
            )

        # shape: (batch_size, beam_size), (batch_size, beam_size)
        (
            start_top_log_probabilities,
            start_predicted_classes,
        ) = start_class_log_probabilities.topk(self.beam_size)
        if self.beam_size == 1 and (start_predicted_classes == self._end_index).all():
            warnings.warn(
                "Empty sequences predicted. You may want to increase the beam size or ensure "
                "your step function is working properly.",
                RuntimeWarning,
            )
            return start_predicted_classes.unsqueeze(-1), start_top_log_probabilities

        # The log probabilities for the last time step.
        # shape: (batch_size, beam_size)
        last_log_probabilities = start_top_log_probabilities

        # shape: [(batch_size, beam_size)]
        predictions.append(start_predicted_classes)

        # Log probability tensor that mandates that the end token is selected.
        # shape: (batch_size * beam_size, num_classes)
        log_probs_after_end = start_class_log_probabilities.new_full(
            (batch_size * self.beam_size, num_classes), float("-inf")
        )
        log_probs_after_end[:, self._end_index] = 0.0

        # Set the same state for each element in the beam.
        for key, state_tensor in state.items():
            if state_tensor is None:
                continue
            _, *last_dims = state_tensor.size()
            # shape: (batch_size * beam_size, *)
            state[key] = (
                state_tensor.unsqueeze(1)
                .expand(batch_size, self.beam_size, *last_dims)
                .reshape(batch_size * self.beam_size, *last_dims)
            )

        for timestep in range(self.max_steps - 1):
            # shape: (batch_size * beam_size,)
            last_predictions = predictions[-1].reshape(batch_size * self.beam_size)

            # If every predicted token from the last step is `self._end_index`,
            # then we can stop early.
            if (last_predictions == self._end_index).all():
                break

            # Take a step. This get the predicted log probs of the next classes
            # and updates the state.
            # shape: (batch_size * beam_size, num_classes)
            class_log_probabilities, state = step(last_predictions, state, timestep + 1)

            # shape: (batch_size * beam_size, num_classes)
            last_predictions_expanded = last_predictions.unsqueeze(-1).expand(
                batch_size * self.beam_size, num_classes
            )

            # Here we are finding any beams where we predicted the end token in
            # the previous timestep and replacing the distribution with a
            # one-hot distribution, forcing the beam to predict the end token
            # this timestep as well.
            # shape: (batch_size * beam_size, num_classes)
            cleaned_log_probabilities = torch.where(
                last_predictions_expanded == self._end_index,
                log_probs_after_end,
                class_log_probabilities,
            )

            # shape (both): (batch_size * beam_size, per_node_beam_size)
            top_log_probabilities, predicted_classes = cleaned_log_probabilities.topk(
                self.per_node_beam_size
            )

            # Here we expand the last log probabilities to (batch_size * beam_size, per_node_beam_size)
            # so that we can add them to the current log probs for this timestep.
            # This lets us maintain the log probability of each element on the beam.
            # shape: (batch_size * beam_size, per_node_beam_size)
            expanded_last_log_probabilities = (
                last_log_probabilities.unsqueeze(2)
                .expand(batch_size, self.beam_size, self.per_node_beam_size)
                .reshape(batch_size * self.beam_size, self.per_node_beam_size)
            )

            # shape: (batch_size * beam_size, per_node_beam_size)
            summed_top_log_probabilities = (
                top_log_probabilities + expanded_last_log_probabilities
            )

            # shape: (batch_size, beam_size * per_node_beam_size)
            reshaped_summed = summed_top_log_probabilities.reshape(
                batch_size, self.beam_size * self.per_node_beam_size
            )

            # shape: (batch_size, beam_size * per_node_beam_size)
            reshaped_predicted_classes = predicted_classes.reshape(
                batch_size, self.beam_size * self.per_node_beam_size
            )

            # Keep only the top `beam_size` beam indices.
            # shape: (batch_size, beam_size), (batch_size, beam_size)
            restricted_beam_log_probs, restricted_beam_indices = reshaped_summed.topk(
                self.beam_size
            )

            # Use the beam indices to extract the corresponding classes.
            # shape: (batch_size, beam_size)
            restricted_predicted_classes = reshaped_predicted_classes.gather(
                1, restricted_beam_indices.long()
            )

            predictions.append(restricted_predicted_classes)

            # shape: (batch_size, beam_size)
            last_log_probabilities = restricted_beam_log_probs

            # The beam indices come from a `beam_size * per_node_beam_size` dimension where the
            # indices with a common ancestor are grouped together. Hence
            # dividing by per_node_beam_size gives the ancestor. (Note that this is integer
            # division as the tensor is a LongTensor.)
            # shape: (batch_size, beam_size)
            backpointer = restricted_beam_indices / self.per_node_beam_size

            backpointers.append(backpointer)

            # Keep only the pieces of the state tensors corresponding to the
            # ancestors created this iteration.
            for key, state_tensor in state.items():
                if state_tensor is None:
                    continue
                _, *last_dims = state_tensor.size()
                # shape: (batch_size, beam_size, *)
                expanded_backpointer = backpointer.view(
                    batch_size, self.beam_size, *([1] * len(last_dims))
                ).expand(batch_size, self.beam_size, *last_dims)

                # shape: (batch_size * beam_size, *)
                state[key] = (
                    state_tensor.reshape(batch_size, self.beam_size, *last_dims)
                    .gather(1, expanded_backpointer.long())
                    .reshape(batch_size * self.beam_size, *last_dims)
                )

            attention_maps.append(state["attention_map"])

        if not torch.isfinite(last_log_probabilities).all():
            warnings.warn(
                "Infinite log probabilities encountered. Some final sequences may not make sense. "
                "This can happen when the beam size is larger than the number of valid (non-zero "
                "probability) transitions that the step function produces.",
                RuntimeWarning,
            )

        reconstructed_predictions = self.reconstruct_sequences(
            predictions, backpointers
        )

        # shape: (batch_size, beam_size, max_steps)
        all_predictions = torch.cat(list(reversed(reconstructed_predictions)), 2)

        return all_predictions, last_log_probabilities, attention_maps


# Image Captioning Data:

# Sources: json (can be of different formats)
# 	- can specify key where image names are present.
# 	- where captions are present.
#   - 'images|images,sentences'
# 	 CSV (image path | captions)
# 	- can specify column names for images.
# 	- column name for captions.
#   - image|captions

# Pytorch Dataset:

# def open_image(im_path):
#     pass
#     # get


class ImageCaptioningDataset(Dataset):
    def __init__(
        self,
        root_path,
        images,
        captions,
        tokenizer,
        vocab,
        transforms,
        chip_sz,
        resize_to,
        norm_stats,
        split,
        flip_vert,
    ):
        # set path and annotations
        self.root_path = Path(root_path)

        # get images
        self.images = images
        self.captions = captions

        # create_vocab using fastai functions and tokenizer or use preloaded vocab.
        self.tokenizer = tokenizer
        self.vocab = vocab

        # if transforms is None Apply default transforms
        self.split = split
        self.resize_to = resize_to
        self.chip_size = chip_sz
        self.transforms = transforms
        self.crop_tfm = crop(size=chip_sz, row_pct=(0, 1), col_pct=(0, 1))
        if transforms is None:
            self.train_transforms, self.val_transforms = get_transforms(
                flip_vert=flip_vert
            )
        elif transforms is False:
            self.train_transforms, self.val_transforms = None, None
        else:
            self.train_transforms, self.val_transforms = transforms

        self.norm_stats = list(map(torch.tensor, norm_stats))

    def __len__(self):
        return len(self.images)

    def __getitem__(self, index):
        # read image using gdal or PIL
        im = open_image(self.root_path / "images" / self.images[index])
        # Apply transforms on the image.

        # get caption using the index.
        cap = self.captions[index]
        cap = cap[random.randint(0, len(cap) - 1)]
        # convert to word index
        cap = self.tokenizer._process_all_1([cap])[0]
        cap = (
            [self.vocab.stoi["xxbos"]]
            + self.vocab.numericalize(cap)
            + [self.vocab.stoi["xxeos"]]
        )
        # return images and captions as indexes

        if len(im.shape) > 2:
            c, h, w = im.shape
        else:
            h, w = im.shape

        if self.chip_size > h or self.chip_size > w:
            crop_flag = False
        else:
            crop_flag = True

        size = (
            (self.resize_to, self.resize_to)
            if self.resize_to is not None
            else self.resize_to
        )
        if self.resize_to is not None:
            self.chip_size = self.resize_to

        if self.transforms is not False:
            if self.split == "train":
                if crop_flag:
                    im = im.apply_tfms(self.crop_tfm).apply_tfms(
                        self.train_transforms, size=size
                    )
                else:
                    im = im.apply_tfms(self.train_transforms, size=size)
            else:
                im = im.resize(self.chip_size)

        return normalize(im.px, *self.norm_stats), cap


def normalize(x, mean, std):
    "Normalize `x` with `mean` and `std`."
    to_tensor = lambda z: torch.tensor(z).to(x.device)
    if type(mean[0]) is not torch.Tensor:
        mean = to_tensor(mean)
        std = to_tensor(std)
    return (x - mean[..., None, None]) / std[..., None, None]


def denormalize(x, mean, std, do_x=True):
    "Denormalize `x` with `mean` and `std`."
    to_tensor = lambda z: torch.tensor(z).to(x.device)
    if type(mean[0]) is not torch.Tensor:
        mean = to_tensor(mean)
        std = to_tensor(std)
    return (
        x.cpu().float() * std[..., None, None] + mean[..., None, None]
        if do_x
        else x.cpu()
    )


def read_json(file_name):
    with open(file_name) as f:
        json_file = json.load(f)
    return json_file


def read_csv(file_name):
    df = pd.read_csv(file_name)
    return df


def parse_json(obj, keys, index, caption_list, club_items):
    if index == (len(keys) - 1):
        if club_items:
            if type(obj) is list:
                caption_list.append([o[keys[index]] for o in obj])
            else:
                caption_list.append([obj[keys[index]]])
        else:
            if type(obj) is list:
                for o in obj:
                    caption_list.append(o[keys[index]])
            else:
                caption_list.append(obj[keys[index]])
    else:
        if type(obj) is list:
            for o in obj:
                parse_json(o[keys[index]], keys, index + 1, caption_list, club_items)
        else:
            obj = obj[keys[index]]
            parse_json(obj, keys, index + 1, caption_list, club_items)


def parse_xml_from_image_file(imagefile):
    """
    Function that returns captions for an image for image captioning data.
    input: imagefile (Path)
    returns: labels (List[str])
    """
    xmlfile = (
        imagefile.parents[1]
        / "labels"
        / imagefile.name.replace("{ims}".format(ims=imagefile.suffix), ".xml")
    )
    labels = ET.parse(xmlfile).getroot().find("object").find("name").text
    return labels


def get_annotations(ann_object, key, read_type):
    if read_type == "json":
        # 'images,filename', 'images,sentences,raw'
        input_key, annotation_key = key.split("|")
        keys = annotation_key.split(",")
        input_keys = input_key.split(",")
        depth = len(keys)
        all_captions = []
        all_images = []
        parse_json(ann_object, keys, 0, all_captions, club_items=True)
        parse_json(ann_object, input_keys, 0, all_images, club_items=False)
        return all_images, all_captions
    if read_type == "csv":
        raise NotImplementedError
    if read_type == "arcgis":
        all_images = ImageList.from_folder(ann_object).items
        all_captions = [[parse_xml_from_image_file(f)] for f in all_images]
        return all_images, all_captions


def collate_fn(data):
    """Creates mini-batch tensors from the list of tuples (image, caption).

    We should build custom collate_fn rather than using default collate_fn,
    because merging caption (including padding) is not supported in default.

    Args:
        data: list of tuple (image, caption).
            - image: torch tensor of shape (3, 256, 256).
            - caption: torch tensor of shape (?); variable length.

    Returns:
        images: torch tensor of shape (batch_size, 3, 256, 256).
        targets: torch tensor of shape (batch_size, padded_length).
        lengths: list; valid length for each padded caption.
    """
    # Sort a data list by caption length (descending order).
    data.sort(key=lambda x: len(x[1]), reverse=True)
    images, captions = zip(*data)

    # Merge images (from tuple of 3D tensor to 4D tensor).
    images = torch.stack(
        [im.px if isinstance(im, fastai.vision.image.Image) else im for im in images], 0
    )

    # Merge captions (from tuple of 1D tensor to 2D tensor).
    lengths = torch.tensor([len(cap) for cap in captions]).long()
    targets = torch.zeros(len(captions), max(lengths)).long()
    for i, cap in enumerate(captions):
        end = lengths[i]
        targets[i, :end] = torch.tensor(cap[:end]).long()
    return images, (targets, lengths)


def prepare_captioning_dataset(
    path, chip_sz, batch_size, val_split_pct, transforms, resize_to, **kwargs
):
    """
    If 'annotations.json' is present in root path, then we should use that,
    also pass 'annotations_key' in wherecaptions are present, seperated by commas
    Elif 'annotations.csv' is present in root path, we should use that.
    pass img col and caption col as seperated by commas
    elif 'labels' folder is present in root path, we should use it. this format is arcgis format
    else format not supported
    """

    if not HAS_FASTAI:
        raise_fastai_import_error(
            import_exception=import_exception, message="", installation_steps=" "
        )

    # Read file and get all captions.
    imagecap_kwargs = kwargs.get("image_captioning_kwargs", {})
    if (path / "annotations.json").exists():
        # RSICD Dataset.
        imagecap_kwargs = {"annotations_key": "images,filename|images,sentences,raw"}
        annotations_json = path / "annotations.json"
        annotations_key = imagecap_kwargs.get("annotations_key")
        annotation_object = read_json(annotations_json)
        ann_type = "json"

    elif (path / "annotations.csv").exists():
        annotations_csv = path / "annotations.csv"
        annotations_key = imagecap_kwargs.get("annotations_key")
        annotation_object = read_csv(annotations_csv)
        ann_type = "csv"

    elif (path / "labels").exists():
        annotation_object = path / "images"
        annotations_key = path / "labels"
        ann_type = "arcgis"

    else:
        raise Exception("Unindentified format")

    lang = imagecap_kwargs.get("language", "en")

    paired_images, paired_captions = get_annotations(
        annotation_object, annotations_key, ann_type
    )

    # create tokenizer and vocab
    tokenizer = Tokenizer(n_cpus=1, lang=lang)
    tokenized_captions = tokenizer.process_all([i for c in paired_captions for i in c])
    vocab = Vocab.create(tokenized_captions, max_vocab=100000, min_freq=4)
    del tokenized_captions

    norm_stats = kwargs.get("norm_stats", imagenet_stats)
    flip_vert = kwargs.get("imagery_type", "satellite") != "oriented"

    # create train val idxs using split pct
    total_files = len(paired_images)
    total_val_files = int(val_split_pct * total_files)
    # random shuffling and maintaining order.
    z = list(zip(paired_images, paired_captions))
    random.shuffle(z)
    paired_images, paired_captions = zip(*z)
    val_images = paired_images[:total_val_files]
    val_captions = paired_captions[:total_val_files]
    train_images = paired_images[total_val_files:]
    train_captions = paired_captions[total_val_files:]

    # instantiate ImageCaptioning class
    train_dataset = ImageCaptioningDataset(
        path,
        train_images,
        train_captions,
        tokenizer,
        vocab,
        transforms,
        chip_sz,
        resize_to,
        norm_stats,
        split="train",
        flip_vert=flip_vert,
    )

    valid_dataset = ImageCaptioningDataset(
        path,
        val_images,
        val_captions,
        tokenizer,
        vocab,
        transforms,
        chip_sz,
        resize_to,
        norm_stats,
        split="val",
        flip_vert=False,
    )

    # create train val dataloaders with appropriate batch size.
    databunch_kwargs = {"num_workers": 0} if sys.platform == "win32" else {}

    train_dl = DataLoader(
        train_dataset, batch_size=batch_size, collate_fn=collate_fn, **databunch_kwargs
    )

    valid_dl = DataLoader(
        valid_dataset, batch_size=batch_size, collate_fn=collate_fn, **databunch_kwargs
    )

    # create Databunch
    device = get_device()
    data = DataBunch(train_dl, valid_dl, device=device, collate_fn=collate_fn)
    # Attach show batch to databunch
    data.show_batch = types.MethodType(show_batch, data)
    # attach vocab
    data.vocab = vocab
    # attach chip size
    data.chip_size = chip_sz
    # attach resize to
    data.resize_to = resize_to
    # add language
    data.lang = lang
    # add path
    data.path = path
    # return databunch.
    data._dataset_type = "ImageCaptioning"
    return data


def show_image_and_text(ax, image, text, show_coords):
    if isinstance(image, fastai.vision.image.Image):
        image = image.px
    ax.imshow(image2np(image))
    # ax.title.set_text('\n'.join(wrap(text, 40)))
    ax.set_title("\n".join(wrap(text, 40)), y=-0.15, pad=0.1)
    ax.title.set_path_effects([PathEffects.withStroke(linewidth=3, foreground="w")])
    if not show_coords:
        ax.axis("off")


def show_batch(self, rows=2, **kwargs):
    # create a grid of square root of rows
    figsize = kwargs.get("figsize", (5 * rows, 5 * rows))
    show_coords = kwargs.get("show_coords", False)
    fig, ax = plt.subplots(rows, rows, figsize=figsize)

    img_idxs = [random.randint(0, len(self.train_ds) - 1) for k in range(rows**2)]
    # iterate through the rows and get transformed images from the dataset class
    for k, idx in enumerate(img_idxs):
        img, captions = self.train_ds[idx]
        caption = self.vocab.textify(captions[1:-1])
        # use matplotlib for display
        show_image_and_text(
            ax[k // rows][k % rows],
            denormalize(img, *self.norm_stats),
            caption,
            show_coords,
        )


def show_results(self, rows, **kwargs):
    figsize = kwargs.get("figsize", (20, rows * 5))
    return_fig = kwargs.get("return_fig", False)
    show_coords = kwargs.get("show_coords", False)
    beam_width = kwargs.get("beam_width", 3)
    max_len = kwargs.get("max_len", 15)
    fig, ax = plt.subplots(rows, 2, figsize=figsize, squeeze=False)
    top = get_top_padding(title_font_size=16, nrows=rows, imsize=5)
    plt.subplots_adjust(top=top)
    fig.suptitle("Ground Truth / Prediction", fontsize=16)
    nbatches = (rows // self._data.valid_dl.batch_size) + 1
    dls = get_nbatches(self._data.valid_dl, nbatches)
    images, captions = [], []
    captions_gts = []

    # concatenating multiple batches after doing predictions.
    for x_in, current_caption_gt in zip(*dls):
        captions_gts.extend(
            [cap.tolist()[: length.item()] for cap, length in zip(*current_caption_gt)]
        )
        current_images, current_captions, _ = self.learn.model.sample(
            x_in, beam_width=beam_width, max_len=max_len
        )
        images.append(current_images)
        captions.extend(current_captions)
    images = torch.cat(images).cpu()
    # iterate through each image and caption print.
    for k, img in enumerate(images):
        # no need to textify predictions because they are already in text format
        caption_pred = captions[k]
        caption_gt = captions_gts[k]
        # textify gt because its still tokenized.
        caption_gt = self._data.vocab.textify(caption_gt[1:-1])
        # use matplotlib for display
        show_image_and_text(
            ax[k][0], denormalize(img, *self._data.norm_stats), caption_gt, show_coords
        )
        show_image_and_text(
            ax[k][1],
            denormalize(img, *self._data.norm_stats),
            caption_pred,
            show_coords,
        )

        if k + 1 == rows:
            break

    # delete empty plots in case of small datasets.
    if k + 1 < rows:
        for i in range(k + 1, rows):
            fig.delaxes(ax[i][0])
            fig.delaxes(ax[i][1])

    if return_fig:
        return fig



import torch
import torch.nn as nn


def conv3x3(in_planes, out_planes, stride=1, groups=1, dilation=1):
    """3x3 convolution with padding"""
    return nn.Conv2d(
        in_planes,
        out_planes,
        kernel_size=3,
        stride=stride,
        padding=dilation,
        groups=groups,
        bias=False,
        dilation=dilation,
    )


def conv1x1(in_planes, out_planes, stride=1):
    """1x1 convolution"""
    return nn.Conv2d(in_planes, out_planes, kernel_size=1, stride=stride, bias=False)


class BasicBlock(nn.Module):
    expansion = 1
    __constants__ = ["downsample"]

    def __init__(
        self,
        inplanes,
        planes,
        stride=1,
        downsample=None,
        groups=1,
        base_width=64,
        dilation=1,
        norm_layer=None,
    ):
        super(BasicBlock, self).__init__()
        if norm_layer is None:
            norm_layer = nn.BatchNorm2d
        if groups != 1 or base_width != 64:
            raise ValueError("BasicBlock only supports groups=1 and base_width=64")
        if dilation > 1:
            raise NotImplementedError("Dilation > 1 not supported in BasicBlock")
        # Both self.conv1 and self.downsample layers downsample the input when stride != 1
        self.conv1 = conv3x3(inplanes, planes, stride)
        self.bn1 = norm_layer(planes)
        self.relu = nn.ReLU(inplace=True)
        self.conv2 = conv3x3(planes, planes)
        self.bn2 = norm_layer(planes)
        self.downsample = downsample
        self.stride = stride

    def forward(self, x):
        identity = x

        out = self.conv1(x)
        out = self.bn1(out)
        out = self.relu(out)

        out = self.conv2(out)
        out = self.bn2(out)

        if self.downsample is not None:
            identity = self.downsample(x)

        out += identity
        out = self.relu(out)

        return out


class Bottleneck(nn.Module):
    expansion = 4
    __constants__ = ["downsample"]

    def __init__(
        self,
        inplanes,
        planes,
        stride=1,
        downsample=None,
        groups=1,
        base_width=64,
        dilation=1,
        norm_layer=None,
    ):
        super(Bottleneck, self).__init__()
        if norm_layer is None:
            norm_layer = nn.BatchNorm2d
        width = int(planes * (base_width / 64.0)) * groups
        # Both self.conv2 and self.downsample layers downsample the input when stride != 1
        self.conv1 = conv1x1(inplanes, width)
        self.bn1 = norm_layer(width)
        self.conv2 = conv3x3(width, width, stride, groups, dilation)
        self.bn2 = norm_layer(width)
        self.conv3 = conv1x1(width, planes * self.expansion)
        self.bn3 = norm_layer(planes * self.expansion)
        self.relu = nn.ReLU(inplace=True)
        self.downsample = downsample
        self.stride = stride

    def forward(self, x):
        identity = x

        out = self.conv1(x)
        out = self.bn1(out)
        out = self.relu(out)

        out = self.conv2(out)
        out = self.bn2(out)
        out = self.relu(out)

        out = self.conv3(out)
        out = self.bn3(out)

        if self.downsample is not None:
            identity = self.downsample(x)

        out += identity
        out = self.relu(out)

        return out


class ResNet(nn.Module):
    def __init__(
        self,
        block,
        layers,
        zero_init_residual=False,
        groups=1,
        widen=1,
        width_per_group=64,
        replace_stride_with_dilation=None,
        norm_layer=None,
        normalize=False,
        output_dim=0,
        hidden_mlp=0,
        nmb_prototypes=0,
        eval_mode=False,
    ):
        super(ResNet, self).__init__()
        if norm_layer is None:
            norm_layer = nn.BatchNorm2d
        self._norm_layer = norm_layer

        self.eval_mode = eval_mode
        self.padding = nn.ConstantPad2d(1, 0.0)

        self.inplanes = width_per_group * widen
        self.dilation = 1
        if replace_stride_with_dilation is None:
            # each element in the tuple indicates if we should replace
            # the 2x2 stride with a dilated convolution instead
            replace_stride_with_dilation = [False, False, False]
        if len(replace_stride_with_dilation) != 3:
            raise ValueError(
                "replace_stride_with_dilation should be None "
                "or a 3-element tuple, got {}".format(replace_stride_with_dilation)
            )
        self.groups = groups
        self.base_width = width_per_group

        # change padding 3 -> 2 compared to original torchvision code because added a padding layer
        num_out_filters = width_per_group * widen
        self.conv1 = nn.Conv2d(
            3, num_out_filters, kernel_size=7, stride=2, padding=2, bias=False
        )
        self.bn1 = norm_layer(num_out_filters)
        self.relu = nn.ReLU(inplace=True)
        self.maxpool = nn.MaxPool2d(kernel_size=3, stride=2, padding=1)
        self.layer1 = self._make_layer(block, num_out_filters, layers[0])
        num_out_filters *= 2
        self.layer2 = self._make_layer(
            block,
            num_out_filters,
            layers[1],
            stride=2,
            dilate=replace_stride_with_dilation[0],
        )
        num_out_filters *= 2
        self.layer3 = self._make_layer(
            block,
            num_out_filters,
            layers[2],
            stride=2,
            dilate=replace_stride_with_dilation[1],
        )
        num_out_filters *= 2
        self.layer4 = self._make_layer(
            block,
            num_out_filters,
            layers[3],
            stride=2,
            dilate=replace_stride_with_dilation[2],
        )
        self.avgpool = nn.AdaptiveAvgPool2d((1, 1))

        # normalize output features
        self.l2norm = normalize

        # projection head
        if output_dim == 0:
            self.projection_head = None
        elif hidden_mlp == 0:
            self.projection_head = nn.Linear(
                num_out_filters * block.expansion, output_dim
            )
        else:
            self.projection_head = nn.Sequential(
                nn.Linear(num_out_filters * block.expansion, hidden_mlp),
                nn.BatchNorm1d(hidden_mlp),
                nn.ReLU(inplace=True),
                nn.Linear(hidden_mlp, output_dim),
            )

        # prototype layer
        self.prototypes = None
        if isinstance(nmb_prototypes, list):
            self.prototypes = MultiPrototypes(output_dim, nmb_prototypes)
        elif nmb_prototypes > 0:
            self.prototypes = nn.Linear(output_dim, nmb_prototypes, bias=False)

        for m in self.modules():
            if isinstance(m, nn.Conv2d):
                nn.init.kaiming_normal_(m.weight, mode="fan_out", nonlinearity="relu")
            elif isinstance(m, (nn.BatchNorm2d, nn.GroupNorm)):
                nn.init.constant_(m.weight, 1)
                nn.init.constant_(m.bias, 0)

        # Zero-initialize the last BN in each residual branch,
        # so that the residual branch starts with zeros, and each residual block behaves like an identity.
        # This improves the model by 0.2~0.3% according to https://arxiv.org/abs/1706.02677
        if zero_init_residual:
            for m in self.modules():
                if isinstance(m, Bottleneck):
                    nn.init.constant_(m.bn3.weight, 0)
                elif isinstance(m, BasicBlock):
                    nn.init.constant_(m.bn2.weight, 0)

    def _make_layer(self, block, planes, blocks, stride=1, dilate=False):
        norm_layer = self._norm_layer
        downsample = None
        previous_dilation = self.dilation
        if dilate:
            self.dilation *= stride
            stride = 1
        if stride != 1 or self.inplanes != planes * block.expansion:
            downsample = nn.Sequential(
                conv1x1(self.inplanes, planes * block.expansion, stride),
                norm_layer(planes * block.expansion),
            )

        layers = []
        layers.append(
            block(
                self.inplanes,
                planes,
                stride,
                downsample,
                self.groups,
                self.base_width,
                previous_dilation,
                norm_layer,
            )
        )
        self.inplanes = planes * block.expansion
        for _ in range(1, blocks):
            layers.append(
                block(
                    self.inplanes,
                    planes,
                    groups=self.groups,
                    base_width=self.base_width,
                    dilation=self.dilation,
                    norm_layer=norm_layer,
                )
            )

        return nn.Sequential(*layers)

    def forward_backbone(self, x):
        x = self.padding(x)

        x = self.conv1(x)
        x = self.bn1(x)
        x = self.relu(x)
        x = self.maxpool(x)
        x = self.layer1(x)
        x = self.layer2(x)
        x = self.layer3(x)
        x = self.layer4(x)

        if self.eval_mode:
            return x

        x = self.avgpool(x)
        x = torch.flatten(x, 1)

        return x

    def forward_head(self, x):
        if self.projection_head is not None:
            x = self.projection_head(x)

        if self.l2norm:
            x = nn.functional.normalize(x, dim=1, p=2)

        if self.prototypes is not None:
            return x, self.prototypes(x)
        return x

    def forward(self, inputs):
        if not isinstance(inputs, list):
            inputs = [inputs]
        idx_crops = torch.cumsum(
            torch.unique_consecutive(
                torch.tensor([inp.shape[-1] for inp in inputs]),
                return_counts=True,
            )[1],
            0,
        )
        start_idx = 0
        for end_idx in idx_crops:
            _out = self.forward_backbone(
                torch.cat(inputs[start_idx:end_idx]).cuda(non_blocking=True)
            )
            if start_idx == 0:
                output = _out
            else:
                output = torch.cat((output, _out))
            start_idx = end_idx
        return self.forward_head(output)


class MultiPrototypes(nn.Module):
    def __init__(self, output_dim, nmb_prototypes):
        super(MultiPrototypes, self).__init__()
        self.nmb_heads = len(nmb_prototypes)
        for i, k in enumerate(nmb_prototypes):
            self.add_module("prototypes" + str(i), nn.Linear(output_dim, k, bias=False))

    def forward(self, x):
        out = []
        for i in range(self.nmb_heads):
            out.append(getattr(self, "prototypes" + str(i))(x))
        return out


def resnet50(**kwargs):
    return ResNet(Bottleneck, [3, 4, 6, 3], **kwargs)


def resnet50w2(**kwargs):
    return ResNet(Bottleneck, [3, 4, 6, 3], widen=2, **kwargs)


def resnet50w4(**kwargs):
    return ResNet(Bottleneck, [3, 4, 6, 3], widen=4, **kwargs)


def resnet50w5(**kwargs):
    return ResNet(Bottleneck, [3, 4, 6, 3], widen=5, **kwargs)



__all__ = ["TST"]


from torch import Tensor
from typing import (
    Iterable,
    Generator,
    Sequence,
    Iterator,
    List,
    Set,
    Dict,
    Union,
    Optional,
)
from arcgis.learn.models._tsmodel_archs.layers import *


class _ScaledDotProductAttention(Module):
    def __init__(self, d_k: int):
        self.d_k = d_k

    def forward(self, q: Tensor, k: Tensor, v: Tensor, mask: Optional[Tensor] = None):
        # MatMul (q, k) - similarity scores for all pairs of positions in an input sequence
        scores = torch.matmul(q, k)  # scores : [bs x n_heads x q_len x q_len]

        # Scale
        scores = scores / (self.d_k**0.5)

        # Mask (optional)
        if mask is not None:
            scores.masked_fill_(mask, -1e9)

        # SoftMax
        attn = F.softmax(scores, dim=-1)  # attn   : [bs x n_heads x q_len x q_len]

        # MatMul (attn, v)
        context = torch.matmul(attn, v)  # context: [bs x n_heads x q_len x d_v]

        return context, attn


class _MultiHeadAttention(Module):
    def __init__(self, d_model: int, n_heads: int, d_k: int, d_v: int):
        r"""
        Input shape:  Q, K, V:[batch_size (bs) x q_len x d_model], mask:[q_len x q_len]
        """
        self.n_heads, self.d_k, self.d_v = n_heads, d_k, d_v

        self.W_Q = nn.Linear(d_model, d_k * n_heads, bias=False)
        self.W_K = nn.Linear(d_model, d_k * n_heads, bias=False)
        self.W_V = nn.Linear(d_model, d_v * n_heads, bias=False)

        self.W_O = nn.Linear(n_heads * d_v, d_model, bias=False)

    def forward(self, Q: Tensor, K: Tensor, V: Tensor, mask: Optional[Tensor] = None):
        bs = Q.size(0)

        # Linear (+ split in multiple heads)
        q_s = (
            self.W_Q(Q).view(bs, -1, self.n_heads, self.d_k).transpose(1, 2)
        )  # q_s    : [bs x n_heads x q_len x d_k]
        k_s = (
            self.W_K(K).view(bs, -1, self.n_heads, self.d_k).permute(0, 2, 3, 1)
        )  # k_s    : [bs x n_heads x d_k x q_len] - transpose(1,2) + transpose(2,3)
        v_s = (
            self.W_V(V).view(bs, -1, self.n_heads, self.d_v).transpose(1, 2)
        )  # v_s    : [bs x n_heads x q_len x d_v]

        # Scaled Dot-Product Attention (multiple heads)
        context, attn = _ScaledDotProductAttention(self.d_k)(
            q_s, k_s, v_s
        )  # context: [bs x n_heads x q_len x d_v], attn: [bs x n_heads x q_len x q_len]

        # Concat
        context = (
            context.transpose(1, 2).contiguous().view(bs, -1, self.n_heads * self.d_v)
        )  # context: [bs x q_len x n_heads * d_v]

        # Linear
        output = self.W_O(context)  # context: [bs x q_len x d_model]

        return output, attn


def get_activation_fn(activation):
    if activation == "relu":
        return nn.ReLU()
    elif activation == "gelu":
        return nn.GELU()
    else:
        return activation()


#         raise ValueError(f'{activation} is not available. You can use "relu" or "gelu"')


def ifnone(a, b):
    return b if a is None else a


class _TSTEncoderLayer(Module):
    def __init__(
        self,
        q_len: int,
        d_model: int,
        n_heads: int,
        d_k: Optional[int] = None,
        d_v: Optional[int] = None,
        d_ff: int = 256,
        dropout: float = 0.1,
        activation: str = "gelu",
    ):
        assert (
            d_model // n_heads
        ), f"d_model ({d_model}) must be divisible by n_heads ({n_heads})"
        d_k = ifnone(d_k, d_model // n_heads)
        d_v = ifnone(d_v, d_model // n_heads)

        # Multi-Head attention
        self.self_attn = _MultiHeadAttention(d_model, n_heads, d_k, d_v)

        # Add & Norm
        self.dropout_attn = nn.Dropout(dropout)
        self.batchnorm_attn = nn.Sequential(
            Transpose(1, 2), nn.BatchNorm1d(d_model), Transpose(1, 2)
        )

        # Position-wise Feed-Forward
        self.ff = nn.Sequential(
            nn.Linear(d_model, d_ff),
            get_activation_fn(activation),
            nn.Dropout(dropout),
            nn.Linear(d_ff, d_model),
        )

        # Add & Norm
        self.dropout_ffn = nn.Dropout(dropout)
        self.batchnorm_ffn = nn.Sequential(
            Transpose(1, 2), nn.BatchNorm1d(d_model), Transpose(1, 2)
        )

    def forward(self, src: Tensor, mask: Optional[Tensor] = None) -> Tensor:
        # Multi-Head attention sublayer
        ## Multi-Head attention
        src2, attn = self.self_attn(src, src, src, mask=mask)
        ## Add & Norm
        src = src + self.dropout_attn(
            src2
        )  # Add: residual connection with residual dropout
        src = self.batchnorm_attn(src)  # Norm: batchnorm

        # Feed-forward sublayer
        ## Position-wise Feed-Forward
        src2 = self.ff(src)
        ## Add & Norm
        src = src + self.dropout_ffn(
            src2
        )  # Add: residual connection with residual dropout
        src = self.batchnorm_ffn(src)  # Norm: batchnorm

        return src


class _TSTEncoder(Module):
    def __init__(
        self,
        q_len,
        d_model,
        n_heads,
        d_k=None,
        d_v=None,
        d_ff=None,
        dropout=0.1,
        activation="gelu",
        n_layers=1,
    ):
        self.layers = nn.ModuleList(
            [
                _TSTEncoderLayer(
                    q_len,
                    d_model,
                    n_heads=n_heads,
                    d_k=d_k,
                    d_v=d_v,
                    d_ff=d_ff,
                    dropout=dropout,
                    activation=activation,
                )
                for i in range(n_layers)
            ]
        )

    def forward(self, src):
        output = src
        for mod in self.layers:
            output = mod(output)
        return output


class TST(Module):
    def __init__(
        self,
        c_in: int,
        c_out: int,
        seq_len: int,
        max_seq_len: Optional[int] = None,
        n_layers: int = 3,
        d_model: int = 128,
        n_heads: int = 16,
        d_k: Optional[int] = None,
        d_v: Optional[int] = None,
        d_ff: int = 256,
        dropout: float = 0.1,
        act: str = "gelu",
        fc_dropout: float = 0.0,
        y_range: Optional[tuple] = None,
        verbose: bool = False,
        **kwargs,
    ):
        r"""TST (Time Series Transformer) is a Transformer that takes continuous time series as inputs.
        As mentioned in the paper, the input must be standardized by_var based on the entire training set.
        Args:
            c_in: the number of features (aka variables, dimensions, channels) in the time series dataset.
            c_out: the number of target classes.
            seq_len: number of time steps in the time series.
            max_seq_len: useful to control the temporal resolution in long time series to avoid memory issues.
            d_model: total dimension of the model (number of features created by the model)
            n_heads:  parallel attention heads.
            d_k: size of the learned linear projection of queries and keys in the MHA. Usual values: 16-512. Default: None -> (d_model/n_heads) = 32.
            d_v: size of the learned linear projection of values in the MHA. Usual values: 16-512. Default: None -> (d_model/n_heads) = 32.
            d_ff: the dimension of the feedforward network model.
            dropout: amount of residual dropout applied in the encoder.
            act: the activation function of intermediate layer, relu or gelu.
            n_layers: the number of sub-encoder-layers in the encoder.
            fc_dropout: dropout applied to the final fully connected layer.
            y_range: range of possible y values (used in regression tasks).
            kwargs: nn.Conv1d kwargs. If not {}, a nn.Conv1d with those kwargs will be applied to original time series.
        Input shape:
            bs (batch size) x nvars (aka features, variables, dimensions, channels) x seq_len (aka time steps)
        """
        self.c_out, self.seq_len = c_out, seq_len

        # Input encoding
        q_len = seq_len
        self.new_q_len = False
        if (
            max_seq_len is not None and seq_len > max_seq_len
        ):  # Control temporal resolution
            self.new_q_len = True
            q_len = max_seq_len
            tr_factor = math.ceil(seq_len / q_len)
            total_padding = tr_factor * q_len - seq_len
            padding = (total_padding // 2, total_padding - total_padding // 2)
            self.W_P = nn.Sequential(
                Pad1d(padding),
                Conv1d(
                    c_in, d_model, kernel_size=tr_factor, padding=0, stride=tr_factor
                ),
            )
            pv(
                f"temporal resolution modified: {seq_len} --> {q_len} time steps: kernel_size={tr_factor}, stride={tr_factor}, padding={padding}.\n",
                verbose,
            )
        elif kwargs:
            self.new_q_len = True
            t = torch.rand(1, 1, seq_len)
            q_len = nn.Conv1d(1, 1, **kwargs)(t).shape[-1]
            self.W_P = nn.Conv1d(c_in, d_model, **kwargs)  # Eq 2
            pv(
                f"Conv1d with kwargs={kwargs} applied to input to create input encodings\n",
                verbose,
            )
        else:
            self.W_P = nn.Linear(
                c_in, d_model
            )  # Eq 1: projection of feature vectors onto a d-dim vector space

        # Positional encoding
        W_pos = torch.empty((q_len, d_model))
        nn.init.uniform_(W_pos, -0.02, 0.02)
        self.W_pos = nn.Parameter(W_pos, requires_grad=True)

        # Residual dropout
        self.dropout = nn.Dropout(dropout)

        # Encoder
        self.encoder = _TSTEncoder(
            q_len,
            d_model,
            n_heads,
            d_k=d_k,
            d_v=d_v,
            d_ff=d_ff,
            dropout=dropout,
            activation=act,
            n_layers=n_layers,
        )
        self.flatten = Flatten()

        # Head
        self.head_nf = q_len * d_model
        self.head = self.create_head(
            self.head_nf, c_out, act=act, fc_dropout=fc_dropout, y_range=y_range
        )

    def create_head(
        self, nf, c_out, act="gelu", fc_dropout=0.0, y_range=None, **kwargs
    ):
        layers = [get_activation_fn(act), Flatten()]
        if fc_dropout:
            layers += [nn.Dropout(fc_dropout)]
        layers += [nn.Linear(nf, c_out)]
        if y_range:
            layers += [SigmoidRange(*y_range)]
        return nn.Sequential(*layers)

    def forward(
        self, x: Tensor, mask: Optional[Tensor] = None
    ) -> Tensor:  # x: [bs x nvars x q_len]
        # Input encoding
        if self.new_q_len:
            u = self.W_P(x).transpose(
                2, 1
            )  # Eq 2        # u: [bs x d_model x q_len] transposed to [bs x q_len x d_model]
        else:
            u = self.W_P(
                x.transpose(2, 1)
            )  # Eq 1                     # u: [bs x q_len x nvars] converted to [bs x q_len x d_model]

        # Positional encoding
        u = self.dropout(u + self.W_pos)

        # Encoder
        z = self.encoder(u)  # z: [bs x q_len x d_model]
        z = z.transpose(2, 1).contiguous()  # z: [bs x d_model x q_len]

        # Classification/ Regression head
        return self.head(z)  # output: [bs x c_out]



import warnings

with warnings.catch_warnings():
    warnings.simplefilter("ignore")
    import tensorflow as tf

    tf.enable_eager_execution()
    import onnx
    from onnx_tf.backend import prepare
import numpy as np
import math


def A(*a):
    return np.array(a[0]) if len(a) == 1 else [np.array(o) for o in a]


imagenet_stats = A([0.485, 0.456, 0.406], [0.229, 0.224, 0.225])

mean = 255 * np.array(imagenet_stats[0], dtype=np.float32)
std = 255 * np.array(imagenet_stats[1], dtype=np.float32)

norm = lambda x: (x - mean) / std
denorm = lambda x: x * std + mean


def pred2dict(bb_np, score, cat_str, c):
    # convert to top left x,y bottom right x,y
    return {
        "x1": bb_np[1],
        "x2": bb_np[3],
        "y1": bb_np[0],
        "y2": bb_np[2],
        "score": score,
        "category": cat_str,
        "class": c,
    }


def to_np(x):
    return np.array(x)


def actn_to_bb(actn, anchors, grid_sizes):
    actn_bbs = tf.tanh(actn).numpy()
    actn_centers = (actn_bbs[:, :2] / 2 * grid_sizes) + anchors[:, :2]
    actn_hw = (actn_bbs[:, 2:] / 2 + 1) * anchors[:, 2:]
    return hw2corners(actn_centers, actn_hw)


def hw2corners(ctr, hw):
    return tf.concat([ctr - hw / 2, ctr + hw / 2], axis=1).numpy()


def nms(boxes, scores, overlap=0.3, top_k=100):
    keep = []
    if len(boxes) == 0:
        return keep
    x1 = boxes[:, 0]
    y1 = boxes[:, 1]
    x2 = boxes[:, 2]
    y2 = boxes[:, 3]

    areas = (x2 - x1) * (y2 - y1)

    order = scores.argsort()
    order = order[-top_k:]

    while order.size > 0:
        i = order[-1]
        keep.append(i)
        order = order[:-1]
        xx1 = np.maximum(x1[i], x1[order])
        yy1 = np.maximum(y1[i], y1[order])
        xx2 = np.minimum(x2[i], x2[order])
        yy2 = np.minimum(y2[i], y2[order])
        w = xx2 - xx1
        h = yy2 - yy1
        inter = w * h
        rem_areas = areas[order]
        union = (rem_areas - inter) + areas[i]
        IoU = inter / union
        order = order[IoU <= overlap]

    return keep, len(keep)


def get_nms_preds(
    bb_class, bb_act, idx, anchors, grid_sizes, classes, nms_overlap, thres
):
    a_ic = actn_to_bb(bb_act[idx], anchors, grid_sizes)
    box_clas = tf.math.sigmoid((bb_class[idx]).max(1)).numpy()
    conf_scores = tf.transpose(tf.math.sigmoid(bb_class[idx])).numpy()

    out_scores, out_box, out_clas = [], [], []

    for i in range(1, len(conf_scores)):
        mask = conf_scores[i] > thres
        if mask.sum() == 0:
            continue
        scores = conf_scores[i][mask]
        mask1 = np.expand_dims(mask, 1) * np.ones(a_ic.shape)
        mask1 = mask1.astype("bool")
        boxes = a_ic[mask1].reshape(-1, 4)
        idx, cnt = nms(boxes, scores, nms_overlap, 50)
        idx = idx[:cnt]
        out_scores.append(scores[idx])
        out_box.append(boxes[idx])
        out_clas.append([i] * cnt)

    if out_clas == []:
        out_clas = [[0]]

    if out_box == []:
        out_box = [[]]

    if out_scores == []:
        out_scores = [[]]

    return predictions(
        out_box[0],
        out_clas[0],
        out_scores[0] if out_scores[0] is not None else None,
        thres,
        classes,
    )


def predictions(bbox, clas=None, prs=None, thresh=0.3, classes=None):
    if len(classes) is None:
        raise Exception("Classes are None")
    else:
        classes = ["bg"] + classes
    bb = bbox
    if prs is None:
        prs = [None] * len(bb)
    if clas is None:
        clas = [None] * len(bb)
    predictions = []
    for i, (b, c, pr) in enumerate(zip(bb, clas, prs)):
        if (b[2] > 0) and (pr is None or pr > thresh):
            cat_str = classes[c]
            score = pr * 100
            bb_np = b.astype("float64")
            predictions.append(pred2dict(bb_np, score, cat_str, c))
    return predictions


def detect_objects_image_space(
    graph, tiles, anchors, grid_sizes, classes, nms_overlap, thres
):
    tile_height, tile_width = tiles.shape[2], tiles.shape[3]

    tiles = tiles.astype("float32")
    img_t = norm(tiles.transpose(0, 2, 3, 1))
    img_t = img_t.transpose(0, 3, 1, 2)

    output_dict = graph.run(img_t)

    clas, bbox = output_dict._0, output_dict._1

    preds = {}

    for batch_idx in range(bbox.shape[0]):
        preds[batch_idx] = get_nms_preds(
            clas, bbox, batch_idx, anchors, grid_sizes, classes, nms_overlap, thres
        )

    batch_size = bbox.shape[0]
    side = math.sqrt(batch_size)

    num_boxes = 0
    for batch_idx in range(batch_size):
        num_boxes = num_boxes + len(preds[batch_idx])

    bounding_boxes = np.empty(shape=(num_boxes, 4), dtype=float)
    scores = np.empty(shape=(num_boxes), dtype=float)
    classes = np.empty(shape=(num_boxes), dtype=np.uint8)

    idx = 0
    for batch_idx in range(batch_size):
        i, j = batch_idx // side, batch_idx % side

        for pred in preds[batch_idx]:
            bounding_boxes[idx, 0] = (pred["y1"] + i) * tile_height
            bounding_boxes[idx, 1] = (pred["x1"] + j) * tile_width
            bounding_boxes[idx, 2] = (pred["y2"] + i) * tile_height
            bounding_boxes[idx, 3] = (pred["x2"] + j) * tile_width
            scores[idx] = pred["score"]
            classes[idx] = pred["class"]
            idx = idx + 1

    return bounding_boxes, scores, classes



from typing import Optional, Dict, Union, List

from arcgis import GIS

from ._feed import Feed
from ._util import _Util
import logging

_LOGGER = logging.getLogger(__name__)


class FeedsManager:
    """
    Used to manage a feed item.

    ==================     ====================================================================
    **Parameter**           **Description**
    ------------------     --------------------------------------------------------------------
    url                    URL of the ArcGIS Velocity organization.
    ------------------     --------------------------------------------------------------------
    gis                    An authenticated :class:`~arcgis.gis.GIS` object.
    ==================     ====================================================================

    """

    _gis = None
    _util = None

    def __init__(self, url: str, gis: GIS):
        self._gis = gis
        self._util = _Util(gis, url)

    @property
    def items(self) -> List[Feed]:
        """
        Get all feeds.

        :return: returns a collection of all configured feed tasks with feed id and feed label.

        .. code-block:: python

            # Get all feeds item

            all_feeds = feeds.items
            all_feeds

        """
        all_feeds_response = self._util._get_request("feeds")
        if all_feeds_response is not None and type(all_feeds_response) is list:
            feed_items = [
                Feed(self._gis, self._util, feed) for feed in all_feeds_response
            ]
            return feed_items
        elif all_feeds_response is None:
            _LOGGER.warning("No Feed items found for the user.")
            return []
        else:
            raise Exception(
                f"Error retrieving Feed items. Velocity response: ${all_feeds_response}"
            )

    def get(self, id) -> Feed:
        """
        Get feed by ID.

        ===============     ====================================================================
        **Parameter**        **Description**
        ---------------     --------------------------------------------------------------------
        id                  Unique ID of a feed.
        ===============     ====================================================================

        :return: endpoint response of feed for the given id and label.

        .. code-block:: python

            # Get feed by id
            # Method: <item>.get(id)

            sample_feed = feeds.get("id")

        """
        feed_item = self._util._get("feed", id)
        return Feed(self._gis, self._util, feed_item)

    def create(self, feed=None) -> Feed:
        """
        Creates a new feed configuration.

        ===============     ====================================================================
        **Parameter**        **Description**
        ---------------     --------------------------------------------------------------------
        feed                An instance of a feed such as RSS or HTTP Poller.
        ===============     ====================================================================

        :return: Id and label of the newly created feed

        .. code-block:: python

            # Usage Example of creating a feature layer feed

            # Connect to a Velocity instance

            from arcgis import GIS
            from arcgis.realtime.velocity.feeds_manager import Feed

            gis = GIS(
                url="https://url.link",
                username="user_name",
                password="user_password",
            )

            velocity = gis.velocity
            feeds = gis.velocity.feeds
            feeds

            # Configure the Feature Layer Feed

            from arcgis.realtime.velocity.feeds import FeatureLayer
            from arcgis.realtime.velocity.http_authentication_type import (
                NoAuth,
                BasicAuth,
                CertificateAuth,
            )

            from arcgis.realtime.velocity.input.format import DelimitedFormat
            from arcgis.realtime.velocity.feeds.geometry import XYZGeometry, SingleFieldGeometry
            from arcgis.realtime.velocity.feeds.time import TimeInterval, TimeInstant
            from arcgis.realtime.velocity.feeds.run_interval import RunInterval

            # feature layer properties

            name = "feature_layer_name"
            description = "feature_layer_description"
            url = "feature_layer_url"
            extent = {
                "spatialReference": {
                    "latestWkid": 3857,
                    "wkid": 102100
                },
                "xmin": "xmin",
                "ymin": "ymin",
                "xmax": "xmax",
                "ymax": "ymax"
            }

            # Set time field

            time = TimeInterval(
                interval_start_field="start_field",
                interval_end_field="end_field"
                # time instant
                # time = TimeInstant(time_field="pubDate")
                # feature_layer_config.set_time_config(time=time)
            )

            # Set recurrence

            run_interval = RunInterval(
                cron_expression="0 * * ? * * *",
                timezone="America/Los_Angeles"
            )

            # Set geometry field - configuring X,Y and Z fields

            geometry = XYZGeometry(
                x_field = "x",
                y_field = "y",
                wkid = 4326
            )

            # a single field geometry could also be configured
            # geometry = SingleFieldFeometry(
                # geometry_field="geometry_field"
                # geometry_type="esriGeometryPoint",
                # geometry_format="esrijson",
                # wkid=4326
            # )
            # feature_layer.set_geometry_config(geometry=geometry)

            feature_layer_config = FeatureLayer(
                label=name,
                description=description,
                query="1=1",
                fields="*",
                outSR=4326,
                url=url,
                extent=extent,
                time_stamp_field=time
            )

            # Manipulate the schema - rename or remove fields, change field data-type

            feature_layer_config.rename_field("org_field_name", "new_field_name")
            feature_layer_config.remove_field("description")

            # Set track id

            feature_layer_config.set_track_id("track_id")

            # Set recurrence

            feature_layer_config.run_interval = RunInterval(
                cron_expression="0 * * ? * * *", timezone="America/Los_Angeles"
            )

            # Create the feed and start it
            feature_layer_feed = feeds.create(feature_layer_config)
            feature_layer_feed.start()
            feeds.items

        """
        if feed is None:
            raise "Feed not found"
        else:
            feed_configuration = feed._build()
            response = self._util._post_request(
                "feed", id=None, payload=feed_configuration
            )
            if response is not None:
                return self.get(response["id"])



"""
Handles security where a user provides the token
"""
from requests.auth import AuthBase
from urllib import parse
from ._schain import SupportMultiAuth
from ..tools import parse_url, assemble_url


class EsriUserTokenAuth(AuthBase, SupportMultiAuth):
    """
    Authentication Using User Created Tokens

    """

    _invalid_token_urls = None
    token = None
    auth = None

    # ----------------------------------------------------------------------
    def __init__(
        self,
        token: str,
        referer: str = None,
        verify_cert: bool = True,
        **kwargs,
    ):
        if token is None:
            raise ValueError("A `token` must be provided")
        self.token = token
        self.verify_cert = verify_cert
        if referer is None:
            self.referer = ""
        else:
            self.referer = referer
        self.legacy = kwargs.pop("legacy", False)

    # ----------------------------------------------------------------------
    def __str__(self):
        return f"<{self.__class__.__name__}, token=.....>"

    # ----------------------------------------------------------------------
    def __repr__(self):
        return f"<{self.__class__.__name__}, token=.....>"

    # ----------------------------------------------------------------------
    def handle_40x(self, r, **kwargs):
        """Handles Case where token is invalid"""

        if (r.status_code < 500 and r.status_code > 399) and str(r.text).lower().find(
            "invalid token"
        ) > -1:
            # Recreate the request without the token
            #
            parsed = parse_url(r.url)
            server_url = assemble_url(parsed)
            self._invalid_token_urls.add(server_url)
            r.content
            r.raw.release_conn()
            r.request.headers.pop("X-Esri-Authorization", None)
            _r = r.connection.send(r.request, **kwargs)
            _r.headers["referer"] = self.referer or "http"
            _r.history.append(r)
            return _r
        return r

    # ----------------------------------------------------------------------
    def __call__(self, r):
        if self._invalid_token_urls is None:
            self._invalid_token_urls = set()
        parsed = parse_url(r.url)
        server_url = assemble_url(parsed)
        if not server_url in self._invalid_token_urls:
            r.register_hook("response", self.handle_40x)
            if self.legacy == False:
                r.headers["X-Esri-Authorization"] = f"Bearer {self.token}"
                r.headers["referer"] = self.referer or ""
            elif self.legacy and r.method == "GET":
                r.prepare_url(url=r.url, params={"token": self.token})
                r.headers["referer"] = self.referer or ""
            elif self.legacy and r.method == "POST":
                data = parse.parse_qs(r.body)
                data["token"] = self.token
                r.prepare_body(data, None, None)
                r.headers["referer"] = self.referer or ""
            else:
                raise Exception(
                    "Only POST and GET are supported with legacy methods of authentication."
                )
            try:
                self.pos = r.body.tell()
            except AttributeError:
                self.pos = None

            return r
        else:
            r.headers.pop("X-Esri-Authorization", None)
        return r
